﻿' (C) Copyright 2011 by Autodesk, Inc. 
'
' Permission to use, copy, modify, and distribute this software
' in object code form for any purpose and without fee is hereby
' granted, provided that the above copyright notice appears in
' all copies and that both that copyright notice and the limited
' warranty and restricted rights notice below appear in all
' supporting documentation.
'
' AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
' AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
' MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK,
' INC. DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL
' BE UNINTERRUPTED OR ERROR FREE.
'
' Use, duplication, or disclosure by the U.S. Government is
' subject to restrictions set forth in FAR 52.227-19 (Commercial
' Computer Software - Restricted Rights) and DFAR 252.227-7013(c)
' (1)(ii)(Rights in Technical Data and Computer Software), as
' applicable.
'
Imports System
Imports System.Collections.Generic
Imports System.Linq

Imports Autodesk.Revit.DB
Imports Autodesk.Revit.DB.Architecture
Imports Autodesk.Revit.UI
Imports Autodesk.Revit.UI.Selection
Imports Autodesk.Revit.ApplicationServices
Imports Autodesk.Revit.Attributes


' Lab7 can copy the picked group to all selected rooms with the 
' * same position. It shows the usage of following basic APIs 
' * 
' *  Room.IsPointIn() to determine if a point is in a room.
' *  Element.BoundingBox to calculate the center of an element.
' *  Pick multiple elements with filter.
' *  
' * Before run this command, a group should  already be in the 
' * active document.
' 


<TransactionAttribute(TransactionMode.Manual)> _
<RegenerationAttribute(RegenerationOption.Manual)> _
Public Class Lab7SelectRoomsAndPlaceGroup
    Implements IExternalCommand
    Public Function Execute(ByVal commandData As ExternalCommandData, _
                          ByRef message As String, _
                          ByVal elements As ElementSet) As Result _
                          Implements IExternalCommand.Execute

        Dim uiApp As UIApplication = commandData.Application
        Dim doc As Document = uiApp.ActiveUIDocument.Document

        Try
            Dim pickedRef As Reference = Nothing
            Dim sel As Selection = uiApp.ActiveUIDocument.Selection

            ' Pick a group
            Dim selFilter As New GroupPickFilter()
            pickedRef = sel.PickObject(ObjectType.Element, _
                                       selFilter, "Please select a group")
            'Get picked Element
            'In Revit 2011 we can get element directly by this line
            Dim elem As Element = pickedRef.Element
            'In Revit 2012, Reference.Element property is obsolete. 
            'The picked element can be retrieved by this:
            'Dim elem as Element = doc.GetElement(pickedRef)

            Dim group As Group = DirectCast(elem, Group)

            ' Get the group's center point
            Dim origin As XYZ = GetElementCenter(group)

            ' Get the room that the picked group is located in
            Dim room As Room = GetRoomOfGroup(doc, origin)

            ' Get the room's center point
            Dim sourceCenter As XYZ = GetRoomCenter(room)

            ' Ask the user to pick target rooms
            Dim roomPickFilter As New RoomPickFilter()
            Dim rooms As IList(Of Reference) = sel.PickObjects( _
              ObjectType.Element, roomPickFilter, _
              "Select target rooms for duplicate furniture group")

            ' Place furniture in each of the rooms
            Dim trans As New Transaction(doc)
            trans.Start("Lab")
            PlaceFurnitureInRooms(doc, rooms, sourceCenter, _
                                  group.GroupType, origin)
            trans.Commit()
            ' If user right-clicks or presess Esc button,
            ' handle the exception
        Catch generatedExceptionName As  _
        Autodesk.Revit.Exceptions.OperationCanceledException
            Return Result.Cancelled
            ' Catch other errors
        Catch ex As Exception
            message = ex.Message
            Return Result.Failed
        End Try

        Return Result.Succeeded
    End Function

    ''' <summary>
    ''' Return the room in which the given point is located
    ''' </summary>
    Private Function GetRoomOfGroup(ByVal doc As Document, _
                                    ByVal point As XYZ) As Room
        Dim collector As New FilteredElementCollector(doc)
        collector.OfCategory(BuiltInCategory.OST_Rooms)
        Dim room As Room = Nothing
        For Each elem As Element In collector
            room = TryCast(elem, Room)
            If room IsNot Nothing Then
                ' Decide if this point is in the picked room                  
                If room.IsPointInRoom(point) Then
                    Exit For
                End If
            End If
        Next
        Return room
    End Function

    ''' <summary>
    ''' Return the center of an element based on its BoundingBox
    ''' </summary>
    Public Function GetElementCenter(ByVal elem As Element) As XYZ
        Dim bounding As BoundingBoxXYZ = elem.BoundingBox(Nothing)
        Dim center As XYZ = (bounding.Max + bounding.Min) * 0.5
        Return center
    End Function

    ''' <summary> 
    ''' Return a room's center point coordinates. 
    ''' Z value is equal to the bottom of the room
    ''' </summary>
    Public Function GetRoomCenter(ByVal room As Room) As XYZ
        ' Get the room center point.
        Dim boundCenter As XYZ = GetElementCenter(room)
        Dim locPt As LocationPoint = DirectCast(room.Location,  _
            LocationPoint)
        Dim roomCenter As New XYZ(boundCenter.X, _
                                  boundCenter.Y, locPt.Point.Z)
        Return roomCenter
    End Function

    ''' <summary>
    ''' Copy the group to each of the provided rooms. The position
    ''' at which the group should be placed is based on the target
    ''' room's center point: it should have the same offset from
    ''' this point as the original had from the center of its room
    ''' </summary>
    Public Sub PlaceFurnitureInRooms( _
      ByVal doc As Document, ByVal rooms As IList(Of Reference), _
      ByVal sourceCenter As XYZ, ByVal gt As GroupType, _
      ByVal groupOrigin As XYZ)

        Dim offset As XYZ = groupOrigin - sourceCenter
        Dim offsetXY As New XYZ(offset.X, offset.Y, 0)

        For Each r As Reference In rooms
            Dim roomTarget As Room = TryCast(r.Element, Room)
            If roomTarget IsNot Nothing Then
                Dim roomCenter As XYZ = GetRoomCenter(roomTarget)
                Dim group As Group = doc.Create.PlaceGroup( _
                roomCenter + offsetXY, gt)
            End If
        Next
    End Sub
End Class

''' <summary>
''' Filter to constrain picking to model groups
''' </summary>
Public Class GroupPickFilter
  Implements ISelectionFilter
  Public Function AllowElement(ByVal e As Element) _
  As Boolean Implements ISelectionFilter.AllowElement
    Return (e.Category.Id.IntegerValue.Equals( _
            CInt(BuiltInCategory.OST_IOSModelGroups)))
  End Function
  Public Function AllowReference( _
  ByVal r As Reference, ByVal p As XYZ) _
  As Boolean Implements ISelectionFilter.AllowReference
    Return False
  End Function
End Class

''' <summary>
''' Filter to constrain picking to rooms
''' </summary>
Public Class RoomPickFilter
  Implements ISelectionFilter
  Public Function AllowElement(ByVal e As Element) _
  As Boolean Implements ISelectionFilter.AllowElement
    Return (e.Category.Id.IntegerValue.Equals( _
            CInt(BuiltInCategory.OST_Rooms)))
  End Function

  Public Function AllowReference(ByVal r As Reference, ByVal p As XYZ) _
  As Boolean Implements ISelectionFilter.AllowReference
    Return False
  End Function
End Class
